from typing import NamedTuple

import jax.random
import tensorflow_probability.substrates.jax as tfp
from jax import numpy as jnp

from jaxns.framework.model import Model
from jaxns.framework.prior import Prior
from jaxns.framework.special_priors import ForcedIdentifiability

tfpd = tfp.distributions


def test_gh144():
    class Output(NamedTuple):
        x: jnp.ndarray
        y: jnp.ndarray

    def prior_model():
        x = yield Prior(dist_or_value=jnp.asarray(0.))
        y = yield Prior(dist_or_value=jnp.asarray(0.))
        return Output(x, y)

    def log_likelihood(z: Output):
        return z.x + z.y

    model = Model(prior_model=prior_model, log_likelihood=log_likelihood)
    assert model.U_ndims == 0
    model.sanity_check(key=jax.random.PRNGKey(0), S=10)


def test_gh95():
    def prior_model():
        x = yield Prior(dist_or_value=jnp.asarray(0.))
        return x

    def log_likelihood(x):
        return jnp.sum(x)

    model = Model(prior_model=prior_model, log_likelihood=log_likelihood)
    assert model.U_ndims == 0


def test_parametrised_singular():
    def prior_model():
        x = yield Prior(dist_or_value=jnp.asarray(0.), name='x').parametrised()
        return x

    def log_likelihood(x):
        return jnp.sum(x)

    _ = Model(prior_model=prior_model, log_likelihood=log_likelihood)


def test_parametrised():
    def prior_model():
        x = yield Prior(tfpd.Uniform(), name='x').parametrised()
        return x

    def log_likelihood(x):
        return x

    model = Model(prior_model=prior_model, log_likelihood=log_likelihood)
    assert model.num_params == 1
    assert model.U_ndims == 0

    log_prob_joint = model.log_prob_joint(model.U_placeholder, allow_nan=True)
    assert log_prob_joint.shape == ()
    assert log_prob_joint == 0.5, "Didn't init at median of uniform"


def test_parametrised_randomised():
    def prior_model():
        x = yield Prior(tfpd.Uniform(), name='x').parametrised(random_init=True)
        y = yield Prior(tfpd.Uniform(), name='y')
        return x, y

    def log_likelihood(x, y):
        return x - y

    model = Model(prior_model=prior_model, log_likelihood=log_likelihood)
    assert model.num_params == 1
    assert model.U_ndims == 1

    log_prob_joint = model.log_prob_joint(model.U_placeholder, allow_nan=True)
    assert log_prob_joint.shape == ()

    model.sanity_check(key=jax.random.PRNGKey(0), S=10)

    model.prepare_input(model.U_placeholder)
    model.transform_parametrised(model.U_placeholder)
    model.forward(model.U_placeholder, allow_nan=True)
    model.sample_U(jax.random.PRNGKey(0))
    model.log_prob_prior(model.U_placeholder)
    model.log_prob_joint(model.U_placeholder, allow_nan=True)
    model.transform(model.U_placeholder)


def test_parametrised_randomised_special():
    def prior_model():
        x = yield ForcedIdentifiability(n=10, low=0., high=1., name='x').parametrised(random_init=True)
        y = yield ForcedIdentifiability(n=10, low=0., high=1., name='y')
        return x, y

    def log_likelihood(x, y):
        return jnp.sum(x - y)

    model = Model(prior_model=prior_model, log_likelihood=log_likelihood)
    assert model.num_params == 10
    assert model.U_ndims == 10

    log_prob_joint = model.log_prob_joint(model.U_placeholder, allow_nan=True)
    assert log_prob_joint.shape == ()

    model.sanity_check(key=jax.random.PRNGKey(0), S=10)

    model.prepare_input(model.U_placeholder)
    model.transform_parametrised(model.U_placeholder)
    model.forward(model.U_placeholder, allow_nan=True)
    model.sample_U(jax.random.PRNGKey(0))
    model.log_prob_prior(model.U_placeholder)
    model.log_prob_joint(model.U_placeholder, allow_nan=True)
    model.transform(model.U_placeholder)


def test_objects_as_likelihood_input():
    class Obj:
        def __init__(self, x):
            self.x = x

    def prior_model():
        x = yield Prior(dist_or_value=jnp.asarray(0.), name='x').parametrised()
        return Obj(x)

    def log_likelihood(obj: Obj):
        return jnp.sum(obj.x)

    model = Model(prior_model=prior_model, log_likelihood=log_likelihood)
    model.sanity_check(key=jax.random.PRNGKey(0), S=10)


def test_empty_prior_models():
    def prior_model():
        return 1.

    def log_likelihood(x):
        return x

    model = Model(prior_model=prior_model, log_likelihood=log_likelihood)
    model.sanity_check(key=jax.random.PRNGKey(0), S=10)



import importlib

import numpy as np


def isinstance_namedtuple(obj) -> bool:
    """
    Check if object is a namedtuple.

    Args:
        obj: object

    Returns:
        bool
    """
    return (
            isinstance(obj, tuple) and
            hasattr(obj, '_asdict') and
            hasattr(obj, '_fields')
    )


def issubclass_namedtuple(cls):
    """
    Check if the type object is a subclass of a namedtuple.
    """
    base_types = cls.__mro__  # Get the method resolution order of the class
    return any(hasattr(base, '_fields') and hasattr(base, '_asdict') for base in base_types)


def serialise_namedtuple(obj):
    if isinstance_namedtuple(obj):
        class_name = f"{obj.__class__.__module__}.{obj.__class__.__name__}"
        return {'type': '__namedtuple__', '__class__': class_name,
                '__data__': {k: serialise_namedtuple(v) for k, v in obj._asdict().items()}}
    elif isinstance(obj, np.ndarray):
        return serialise_ndarray(obj)
    elif isinstance(obj, (list, tuple)):
        return [serialise_namedtuple(v) for v in obj]
    elif isinstance(obj, dict):
        return {k: serialise_namedtuple(v) for k, v in obj.items()}
    else:
        return obj


def deserialise_namedtuple(obj):
    if isinstance(obj, dict) and 'type' in obj and obj['type'] == '__namedtuple__':
        class_path = obj['__class__']
        module_name, class_name = class_path.rsplit('.', 1)
        module = importlib.import_module(module_name)
        class_ = getattr(module, class_name)
        return class_(**{k: deserialise_namedtuple(v) for k, v in obj['__data__'].items()})
    elif isinstance(obj, dict) and 'type' in obj and obj['type'] == '__ndarray__':
        return deserialise_ndarray(obj)
    elif isinstance(obj, (list, tuple)):
        return [deserialise_namedtuple(v) for v in obj]
    elif isinstance(obj, dict):
        return {k: deserialise_namedtuple(v) for k, v in obj.items()}
    return obj


def serialise_ndarray(obj):
    if isinstance(obj, np.ndarray):
        return {'type': '__ndarray__', '__dtype__': str(obj.dtype), '__data__': obj.tolist()}
    return obj


def deserialise_ndarray(obj):
    if isinstance(obj, dict) and obj.get('type') == '__ndarray__':
        return np.array(obj['__data__'], dtype=obj['__dtype__'])
    return obj



import jax
import numpy as np
import pytest
from jax import numpy as jnp

from jaxns.internals.cumulative_ops import cumulative_op_static, cumulative_op_dynamic, scan_associative_cumulative_op
from jaxns.internals.types import float_type, int_type


def test_cumulative_op_static():
    def op(accumulate, y):
        return accumulate + y

    init = jnp.asarray(0, float_type)
    xs = jnp.asarray([1, 2, 3], float_type)
    final_accumulate, result = cumulative_op_static(op=op, init=init, xs=xs)
    assert final_accumulate == 6
    assert all(result == jnp.asarray([1, 3, 6], float_type))

    final_accumulate, result = cumulative_op_static(op=op, init=init, xs=xs, pre_op=True)
    assert final_accumulate == 6
    assert all(result == jnp.asarray([0, 1, 3], float_type))


@pytest.mark.parametrize("binary_op", [jnp.add, jnp.multiply, jnp.minimum, jnp.maximum])
def test_scan_associative_cumulative_op(binary_op):
    def op(accumulate, y):
        return binary_op(accumulate, y)

    init = jnp.asarray(1, float_type)
    xs = jnp.arange(1, 11, dtype=float_type)
    final_accumulate, result = scan_associative_cumulative_op(op=binary_op, init=init, xs=xs)
    final_accumulate_expected, result_expected = cumulative_op_static(op=op, init=init, xs=xs)
    assert final_accumulate == final_accumulate_expected
    np.testing.assert_allclose(result, result_expected)

    final_accumulate, result = scan_associative_cumulative_op(op=op, init=init, xs=xs, pre_op=True)
    final_accumulate_expected, result_expected = cumulative_op_static(op=op, init=init, xs=xs, pre_op=True)
    assert final_accumulate == final_accumulate_expected
    np.testing.assert_allclose(result, result_expected)


@pytest.mark.parametrize("binary_op", [jnp.subtract, jnp.true_divide])
def test_scan_associative_cumulative_op(binary_op):
    def op(accumulate, y):
        return binary_op(accumulate, y)

    init = jnp.asarray(1, float_type)
    xs = jnp.arange(1, 11, dtype=float_type)
    final_accumulate, result = scan_associative_cumulative_op(op=binary_op, init=init, xs=xs)
    final_accumulate_expected, result_expected = cumulative_op_static(op=op, init=init, xs=xs)
    with pytest.raises(AssertionError):
        assert final_accumulate == final_accumulate_expected
    with pytest.raises(AssertionError):
        np.testing.assert_allclose(result, result_expected)

    final_accumulate, result = scan_associative_cumulative_op(op=op, init=init, xs=xs, pre_op=True)
    final_accumulate_expected, result_expected = cumulative_op_static(op=op, init=init, xs=xs, pre_op=True)
    with pytest.raises(AssertionError):
        assert final_accumulate == final_accumulate_expected
    with pytest.raises(AssertionError):
        np.testing.assert_allclose(result, result_expected)


def test_scan_associative_cumulative_op_with_pytrees():
    # Test with pytrees for xs and ys
    def op(accumulate, y):
        return jax.tree.map(lambda x, y: jnp.add(x, y), accumulate, y)

    init = {'a': jnp.asarray(0, float_type), 'b': jnp.asarray(0, float_type)}
    xs = {'a': jnp.asarray([1, 2, 3], float_type), 'b': jnp.asarray([4, 5, 6], float_type)}
    final_accumulate, result = scan_associative_cumulative_op(op=op, init=init, xs=xs)
    assert final_accumulate == {'a': 6, 'b': 15}
    assert all(result['a'] == jnp.asarray([1, 3, 6], float_type))
    assert all(result['b'] == jnp.asarray([4, 9, 15], float_type))

    final_accumulate, result = scan_associative_cumulative_op(op=op, init=init, xs=xs, pre_op=True)
    assert final_accumulate == {'a': 6, 'b': 15}
    assert all(result['a'] == jnp.asarray([0, 1, 3], float_type))
    assert all(result['b'] == jnp.asarray([0, 4, 9], float_type))


def test_cumulative_op_dynamic():
    def op(accumulate, y):
        return accumulate + y

    init = jnp.asarray(0, float_type)
    xs = jnp.asarray([1, 2, 3], float_type)
    stop_idx = jnp.asarray(3, int_type)
    final_accumulate, result = cumulative_op_dynamic(op=op, init=init, xs=xs, stop_idx=stop_idx)
    assert final_accumulate == 6
    assert all(result == jnp.asarray([1, 3, 6], float_type))

    final_accumulate, result = cumulative_op_dynamic(op=op, init=init, xs=xs, stop_idx=stop_idx, pre_op=True)
    assert final_accumulate == 6
    assert all(result == jnp.asarray([0, 1, 3], float_type))

    stop_idx = jnp.asarray(2, int_type)
    final_accumulate, result = cumulative_op_dynamic(op=op, init=init, xs=xs, stop_idx=stop_idx)
    assert final_accumulate == 3
    assert all(result == jnp.asarray([1, 3, 0], float_type))

    final_accumulate, result = cumulative_op_dynamic(op=op, init=init, xs=xs, stop_idx=stop_idx, pre_op=True)
    assert final_accumulate == 3
    assert all(result == jnp.asarray([0, 1, 0], float_type))



import warnings
from typing import List, Tuple, Union

import jax.random
import numpy as np
import pytest
import tensorflow_probability.substrates.jax as tfp
from jax import random, numpy as jnp, vmap

from jaxns.framework.bases import PriorModelGen, BaseAbstractPrior
from jaxns.framework.ops import parse_prior, prepare_input, compute_log_likelihood
from jaxns.framework.prior import Prior, InvalidPriorName
from jaxns.framework.special_priors import Bernoulli, Categorical, Poisson, Beta, ForcedIdentifiability, \
    UnnormalisedDirichlet, _poisson_quantile_bisection, _poisson_quantile, Empirical, TruncationWrapper
from jaxns.framework.wrapped_tfp_distribution import InvalidDistribution, distribution_chain
from jaxns.internals.types import float_type

tfpd = tfp.distributions


def test_single_prior():
    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=2), name='x')
        y = yield Prior(tfpd.Normal(loc=2, scale=x), name='x')
        z = x + y
        return z

    try:
        parse_prior(prior_model)
        assert False
    except InvalidPriorName:
        assert True

    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=2), name='x')
        y = yield Prior(tfpd.Normal(loc=2, scale=x), name='y')
        z = x + y
        return z

    U, X = parse_prior(prior_model)
    U_ndims = U.shape[0]
    assert isinstance(X, dict)
    assert len(X) == 2
    assert U_ndims == 2

    U = random.uniform(random.PRNGKey(42), shape=(U_ndims,), dtype=float_type)

    def log_likelihood(Z):
        return Z - jnp.sum(Z ** 2)

    log_L = compute_log_likelihood(U, prior_model, log_likelihood=log_likelihood)
    assert not jnp.isnan(log_L)


def test_no_quantile_prior():
    def prior_model() -> PriorModelGen:
        z = yield Prior(tfpd.VonMises(loc=0., concentration=1.))
        return z

    try:
        parse_prior(prior_model)
        assert False
    except InvalidDistribution:
        assert True


def test_prior_model_basic():
    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=2))
        y = yield Prior(tfpd.Normal(loc=2, scale=x))
        z = x + y
        return z, z ** 2

    U, X = parse_prior(prior_model)
    U_ndims = U.shape[0]
    assert isinstance(X, dict)
    assert len(X) == 0
    assert U_ndims == 2

    U = random.uniform(random.PRNGKey(42), shape=(U_ndims,), dtype=float_type)
    output = prepare_input(U, prior_model)
    assert isinstance(output, tuple)
    assert len(output) == 2
    assert jnp.allclose(output[1], output[0] ** 2)

    def log_likelihood(Z, Z2):
        return Z - jnp.sum(Z2)

    log_L = compute_log_likelihood(U, prior_model, log_likelihood=log_likelihood)
    assert not jnp.isnan(log_L)


def test_distribution_chain():
    d = tfpd.MultivariateNormalTriL(loc=jnp.zeros(5), scale_tril=jnp.eye(5))
    chain = distribution_chain(d)
    assert len(chain) == 2
    assert isinstance(chain[0], tfpd.Sample)
    assert isinstance(chain[1], tfpd.MultivariateNormalTriL)

    chain = distribution_chain(tfpd.Normal(loc=jnp.zeros(5), scale=jnp.ones(5)))
    assert len(chain) == 1
    assert isinstance(chain[0], tfpd.Normal)


def test_priors():
    d = Prior(jnp.zeros(5))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (0,)
    assert d.shape == (5,)

    d = Prior(tfpd.Uniform(low=jnp.zeros(5), high=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.Normal(loc=jnp.zeros(5), scale=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.Laplace(loc=jnp.zeros(5), scale=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.Cauchy(loc=jnp.zeros(5), scale=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.StudentT(df=1.5, loc=jnp.zeros(5), scale=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.Beta(concentration0=jnp.ones(5), concentration1=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.HalfNormal(scale=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.HalfCauchy(loc=jnp.zeros(5), scale=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.Gamma(concentration=jnp.ones(5), rate=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)

    d = Prior(tfpd.Gumbel(loc=jnp.ones(5), scale=jnp.ones(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == ()
    assert d.shape == (5,)

    d = Prior(tfpd.MultivariateNormalTriL(loc=jnp.zeros(5), scale_tril=jnp.eye(5)))
    print(d)
    assert d.forward(jnp.ones(d.base_shape, float_type)).shape == d.shape
    assert d.forward(jnp.zeros(d.base_shape, float_type)).shape == d.shape
    assert d.base_shape == (5,)
    assert d.shape == (5,)


@pytest.fixture(scope='module')
def mock_special_priors() -> List[
    Tuple[BaseAbstractPrior, Tuple[Union[float, int], Union[float, int]], Tuple[int, ...]]]:
    # prior, (min, max), shape
    return [
        (Bernoulli(probs=jnp.ones(5), name='x'), (0, 1), (5,)),
        (Categorical(parametrisation='gumbel_max', probs=jnp.ones(5), name='x'), (0, 4), ()),
        (Categorical(parametrisation='cdf', probs=jnp.ones(5), name='x'), (0, 4), ()),
        (Poisson(rate=jnp.ones(5), name='x'), (0, 100), (5,)),
        (Beta(concentration0=jnp.ones(5), concentration1=jnp.ones(5), name='x'), (0, 1), (5,)),
        (ForcedIdentifiability(n=10, low=jnp.zeros(5), high=jnp.ones(5), name='x'), (0, 1), (10, 5)),
        (ForcedIdentifiability(n=10, low=jnp.zeros(5), high=jnp.ones(5), fix_left=True, name='x'), (0, 1), (10, 5)),
        (ForcedIdentifiability(n=10, low=jnp.zeros(5), high=jnp.ones(5), fix_right=True, name='x'), (0, 1), (10, 5)),
        (ForcedIdentifiability(n=10, low=jnp.zeros(5), high=jnp.ones(5), fix_left=True, fix_right=True, name='x'),
         (0, 1), (10, 5)),
        (UnnormalisedDirichlet(concentration=jnp.ones(5), name='x'), (0, jnp.inf), (5,)),
        (UnnormalisedDirichlet(concentration=jnp.ones((3, 5)), name='x'), (0, jnp.inf), (3, 5)),
    ]


def test_special_priors(mock_special_priors: List[BaseAbstractPrior]):
    for prior, (vmin, vmax), shape in mock_special_priors:
        print(f"Testing {prior.__class__}")
        x = prior.forward(jnp.ones(prior.base_shape, float_type))
        assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
        assert jnp.all(x >= vmin)
        assert jnp.all(x <= vmax)
        assert x.shape == shape
        assert x.shape == prior.shape
        x = prior.forward(jnp.zeros(prior.base_shape, float_type))
        assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
        assert jnp.all(x >= vmin)
        assert jnp.all(x <= vmax)
        assert x.shape == shape
        assert x.shape == prior.shape

        u_input = vmap(lambda key: random.uniform(key, shape=prior.base_shape))(random.split(random.PRNGKey(42), 10))
        x = vmap(lambda u: prior.forward(u))(u_input)
        assert jnp.all(x >= vmin)
        assert jnp.all(x <= vmax)

        try:
            u = vmap(lambda x: prior.inverse(x))(x)
            assert u.shape[1:] == prior.base_shape

            if prior.dtype in [jnp.bool_, jnp.int32, jnp.int64]:
                continue
            print(u)
            print(u_input)
            assert jnp.allclose(u, u_input)
        except NotImplementedError:
            warnings.warn(f"Skipping inverse test for {prior.__class__}")
            pass


@pytest.mark.parametrize("rate, error", (
        [2.0, 1.],
        [10., 1.],
        [100., 1.],
        [1000., 1.],
        [10000., 1.]
)
                         )
def test_poisson_quantile_bisection(rate, error):
    U = jnp.linspace(0., 1. - np.spacing(1.), 1000)
    x, x_results = _poisson_quantile_bisection(U, rate, unroll=False)
    diff_last_two = jnp.abs(x_results[..., -1] - x_results[..., -2])

    # Make sure less than 1 apart
    assert jnp.all(diff_last_two <= error)


@pytest.mark.parametrize("rate", [2.0, 10., 100., 1000., 10000.])
def test_poisson_quantile(rate):
    U = jnp.linspace(0., 1. - np.spacing(1.), 10000)
    x = _poisson_quantile(U, rate)
    assert jnp.all(jnp.isfinite(x))


def test_forced_identifiability():
    x = jnp.asarray([0., 0.1, 0.2, 1.])
    prior = ForcedIdentifiability(n=4, low=0., high=1., name='x')
    u = prior.inverse(x)
    print(u)
    assert jnp.allclose(prior.forward(u), x)

    x = jnp.asarray([0., 0.11, 0.1, 1.])
    prior = ForcedIdentifiability(n=4, low=0., high=1., name='x')
    u = prior.inverse(x)
    print(u)
    print(prior.forward(u))
    assert jnp.allclose(prior.forward(u), x)

    x = jnp.asarray([0., 0.1, 0.2, 1.])
    prior = ForcedIdentifiability(n=4, low=0., high=1., fix_right=True, name='x')
    u = prior.inverse(x)
    print(u)
    assert jnp.allclose(prior.forward(u), x)

    x = jnp.asarray([0., 0.11, 0.1, 1.])
    prior = ForcedIdentifiability(n=4, low=0., high=1., fix_right=True, name='x')
    u = prior.inverse(x)
    print(u)
    print(prior.forward(u))
    assert jnp.allclose(prior.forward(u), x)

    x = jnp.asarray([0., 0.1, 0.2, 1.])
    prior = ForcedIdentifiability(n=4, low=0., high=1., fix_left=True, name='x')
    u = prior.inverse(x)
    print(u)
    assert jnp.allclose(prior.forward(u), x)

    x = jnp.asarray([0., 0.11, 0.11, 1.])
    prior = ForcedIdentifiability(n=4, low=0., high=1., fix_left=True, name='x')
    u = prior.inverse(x)
    print(u)
    print(prior.forward(u))
    assert jnp.allclose(prior.forward(u), x)

    x = jnp.asarray([0., 0.1, 0.2, 1.])
    prior = ForcedIdentifiability(n=4, low=0., high=1., fix_left=True, fix_right=True, name='x')
    u = prior.inverse(x)
    print(u)
    assert jnp.allclose(prior.forward(u), x)

    x = jnp.asarray([0., 0.11, 0.1, 1.])
    prior = ForcedIdentifiability(n=4, low=0., high=1., fix_left=True, fix_right=True, name='x')
    u = prior.inverse(x)
    print(u)
    print(prior.forward(u))
    assert jnp.allclose(prior.forward(u), x)


def test_empirical():
    samples = jax.random.normal(jax.random.PRNGKey(42), shape=(5, 1000))
    prior = Empirical(samples=samples, resolution=100, name='x')
    assert prior._percentiles.shape == (101, 5)

    x = prior.forward(jnp.ones(prior.base_shape, float_type))
    assert x.shape == (5,)
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    x = prior.forward(jnp.zeros(prior.base_shape, float_type))
    assert x.shape == (5,)
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))

    x = prior.forward(0.5 * jnp.ones(prior.base_shape, float_type))
    assert jnp.allclose(x, 0., atol=0.05)

    u_input = vmap(lambda key: random.uniform(key, shape=prior.base_shape))(random.split(random.PRNGKey(42), 1000))
    x = vmap(lambda u: prior.forward(u))(u_input)
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))

    u = vmap(lambda x: prior.inverse(x))(x)
    # print(u)
    assert jnp.allclose(u, u_input)
    assert u.shape[1:] == prior.base_shape

def test_truncation_wrapper():
    prior = Prior(tfpd.Normal(loc=jnp.zeros(5), scale=jnp.ones(5)))
    trancated_prior = TruncationWrapper(prior=prior, low=0., high=1.)

    x = trancated_prior.forward(jnp.ones(trancated_prior.base_shape, float_type))
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    assert jnp.all(x >= 0.)
    assert jnp.all(x <= 1.)
    assert x.shape == (5,)

    x = trancated_prior.forward(jnp.zeros(trancated_prior.base_shape, float_type))
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    assert jnp.all(x >= 0.)
    assert jnp.all(x <= 1.)
    assert x.shape == (5,)

    u_input = vmap(lambda key: random.uniform(key, shape=trancated_prior.base_shape))(random.split(random.PRNGKey(42), 1000))
    x = vmap(lambda u: trancated_prior.forward(u))(u_input)
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    assert jnp.all(x >= 0.)
    assert jnp.all(x <= 1.)

    u = vmap(lambda x: trancated_prior.inverse(x))(x)
    np.testing.assert_allclose(u, u_input, atol=5e-7)

    prior = Prior(tfpd.Normal(loc=jnp.zeros(5), scale=jnp.ones(5)))
    trancated_prior = TruncationWrapper(prior=prior, low=-jnp.inf, high=1.)

    x = trancated_prior.forward(jnp.ones(trancated_prior.base_shape, float_type))
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    assert jnp.all(x >= -jnp.inf)
    assert jnp.all(x <= 1.)
    assert x.shape == (5,)

    x = trancated_prior.forward(jnp.zeros(trancated_prior.base_shape, float_type))
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    assert jnp.all(x >= -jnp.inf)
    assert jnp.all(x <= 1.)
    assert x.shape == (5,)

    u_input = vmap(lambda key: random.uniform(key, shape=trancated_prior.base_shape))(
        random.split(random.PRNGKey(42), 1000))
    x = vmap(lambda u: trancated_prior.forward(u))(u_input)
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    assert jnp.all(x >= -jnp.inf)
    assert jnp.all(x <= 1.)

    u = vmap(lambda x: trancated_prior.inverse(x))(x)
    np.testing.assert_allclose(u, u_input, atol=5e-7)

    prior = Prior(tfpd.Normal(loc=jnp.zeros(5), scale=0.01*jnp.ones(5)))
    trancated_prior = TruncationWrapper(prior=prior, low=0., high=1.)

    x = trancated_prior.forward(jnp.ones(trancated_prior.base_shape, float_type))
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    assert jnp.all(x >= 0.)
    assert jnp.all(x <= 1.)
    assert x.shape == (5,)

    x = trancated_prior.forward(jnp.zeros(trancated_prior.base_shape, float_type))
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    assert jnp.all(x >= 0.)
    assert jnp.all(x <= 1.)
    assert x.shape == (5,)

    u_input = vmap(lambda key: random.uniform(key, shape=trancated_prior.base_shape))(
        random.split(random.PRNGKey(42), 1000))
    x = vmap(lambda u: trancated_prior.forward(u))(u_input)
    assert jnp.all(jnp.bitwise_not(jnp.isnan(x)))
    assert jnp.all(x >= 0.)
    assert jnp.all(x <= 1.)

    u = vmap(lambda x: trancated_prior.inverse(x))(x)
    np.testing.assert_allclose(u, u_input, atol=5e-7)



from jaxns.framework.model import *
from jaxns.framework.prior import *
from jaxns.framework.special_priors import *
from jaxns.framework.jaxify import *
from jaxns.framework.bases import PriorModelGen, PriorModelType


from jaxns.internals.types import TerminationConditionDisjunction, TerminationConditionConjunction, \
    TerminationCondition, isinstance_namedtuple


def test_termination_condition():
    assert isinstance(TerminationCondition() | TerminationCondition(), TerminationConditionDisjunction)
    assert isinstance(TerminationCondition() & TerminationCondition(), TerminationConditionConjunction)


def test_isinstance_namedtuple():
    from collections import namedtuple
    Point = namedtuple('Point', ['x', 'y'])
    p = Point(1, 2)
    assert isinstance_namedtuple(p)

    # Test that it works with a custom namedtuple
    class Point2(namedtuple('Point2', ['x', 'y'])):
        pass

    p2 = Point2(1, 2)

    assert isinstance_namedtuple(p2)

    assert not isinstance_namedtuple((1, 2))



import io
from typing import NamedTuple, Optional, Union, TextIO, Tuple, List

import jax
import jax.nn
import jax.numpy as jnp
import numpy as np
from jax import lax, random, pmap
from jax._src.lax import parallel
from jax._src.scipy.special import logit
from jaxopt import NonlinearCG

from jaxns.framework.bases import BaseAbstractModel
from jaxns.internals.logging import logger
from jaxns.internals.maps import remove_chunk_dim
from jaxns.internals.types import PRNGKey, StaticStandardNestedSamplerState, BoolArray, StaticStandardSampleCollection, \
    int_type, Sample, IntArray, UType, FloatArray, LikelihoodInputType, XType
from jaxns.nested_sampler.standard_static import draw_uniform_samples, _inter_sync_shrinkage_process, \
    create_init_termination_register
from jaxns.samplers.bases import BaseAbstractSampler
from jaxns.utils import _bit_mask

__all__ = [
    'GlobalOptimisationResults',
    'GlobalOptimisationTerminationCondition',
    'GlobalOptimisationState',
    'SimpleGlobalOptimisation'
]


class GlobalOptimisationState(NamedTuple):
    key: PRNGKey
    samples: Sample
    num_likelihood_evaluations: IntArray
    num_samples: IntArray


class GlobalOptimisationResults(NamedTuple):
    U_solution: UType
    X_solution: XType
    solution: LikelihoodInputType
    log_L_solution: FloatArray
    num_likelihood_evaluations: IntArray
    num_samples: IntArray
    termination_reason: IntArray
    relative_spread: FloatArray
    absolute_spread: FloatArray


class GlobalOptimisationTerminationCondition(NamedTuple):
    max_likelihood_evaluations: Optional[Union[IntArray, int]] = None  # jnp.asarray(jnp.iinfo(int_type).max, int_type)
    log_likelihood_contour: Optional[
        Union[FloatArray, float]] = None  # jnp.asarray(jnp.finfo(float_type).max, float_type)
    rtol: Optional[Union[FloatArray, float]] = None  # jnp.asarray(0., float_type)
    atol: Optional[Union[FloatArray, float]] = None  # jnp.asarray(0., float_type)
    min_efficiency: Optional[Union[FloatArray, float]] = None  # jnp.asarray(0., float_type)

    def __and__(self, other):
        return TerminationConditionConjunction(conds=[self, other])

    def __or__(self, other):
        return TerminationConditionDisjunction(conds=[self, other])


class TerminationConditionConjunction(NamedTuple):
    conds: List[Union[
        'TerminationConditionDisjunction', 'TerminationConditionConjunction', GlobalOptimisationTerminationCondition]]


class TerminationConditionDisjunction(NamedTuple):
    conds: List[Union[
        'TerminationConditionDisjunction', TerminationConditionConjunction, GlobalOptimisationTerminationCondition]]


def determine_termination(term_cond: GlobalOptimisationTerminationCondition,
                          state: GlobalOptimisationState) -> Tuple[BoolArray, IntArray]:
    """
    Determine if termination should happen. Termination Flags are bits:
        0-bit -> 1: used maximum allowed number of likelihood evaluations
        1-bit -> 2: reached goal log-likelihood contour
        2-bit -> 4: relative spread of log-likelihood values below threshold
        3-bit -> 8: absolute spread of log-likelihood values below threshold
        4-bit -> 16: efficiency below threshold
        5-bit -> 32: on a plateau (possibly local minimum, or due to numerical issues)

    Multiple flags are summed together

    Args:
        term_cond: termination condition
        state: global optimisation state

    Returns:
        boolean done signal, and termination reason
    """

    termination_reason = jnp.asarray(0, int_type)
    done = jnp.asarray(False, jnp.bool_)

    def _set_done_bit(bit_done, bit_reason, done, termination_reason):
        if bit_done.size > 1:
            raise RuntimeError("bit_done must be a scalar.")
        done = jnp.bitwise_or(bit_done, done)
        termination_reason += jnp.where(bit_done,
                                        jnp.asarray(2 ** bit_reason, int_type),
                                        jnp.asarray(0, int_type))
        return done, termination_reason

    if isinstance(term_cond, TerminationConditionConjunction):
        for c in term_cond.conds:
            _done, _reason = determine_termination(term_cond=c, state=state)
            done = jnp.bitwise_and(_done, done)
            termination_reason = jnp.bitwise_and(_reason, termination_reason)
        return done, termination_reason

    if isinstance(term_cond, TerminationConditionDisjunction):
        for c in term_cond.conds:
            _done, _reason = determine_termination(term_cond=c, state=state)
            done = jnp.bitwise_or(_done, done)
            termination_reason = jnp.bitwise_or(_reason, termination_reason)
        return done, termination_reason

    if term_cond.max_likelihood_evaluations is not None:
        # used all points
        reached_max_samples = state.num_likelihood_evaluations >= term_cond.max_likelihood_evaluations
        done, termination_reason = _set_done_bit(reached_max_samples, 0,
                                                 done=done, termination_reason=termination_reason)

    if term_cond.log_likelihood_contour is not None:
        # reached goal log-likelihood contour
        reached_log_L_contour = jnp.max(state.samples.log_L) >= term_cond.log_likelihood_contour
        done, termination_reason = _set_done_bit(reached_log_L_contour, 1,
                                                 done=done, termination_reason=termination_reason)

    if term_cond.rtol is not None:
        # relative spread of log-likelihood values below threshold
        max_log_L = jnp.max(state.samples.log_L)
        min_log_L = jnp.min(state.samples.log_L)
        diff_log_L = jnp.abs(max_log_L - min_log_L)  # NaN = inf - inf
        diff_log_L = jnp.where(jnp.isnan(diff_log_L), jnp.inf, diff_log_L)
        mean_log_L = 0.5 * jnp.abs(max_log_L + min_log_L)  # = inf - inf
        mean_log_L = jnp.where(jnp.isnan(mean_log_L), jnp.inf, mean_log_L)
        reached_rtol = diff_log_L <= term_cond.rtol * mean_log_L
        done, termination_reason = _set_done_bit(reached_rtol, 2,
                                                 done=done, termination_reason=termination_reason)

    if term_cond.atol is not None:
        # absolute spread of log-likelihood values below threshold
        max_log_L = jnp.max(state.samples.log_L)
        min_log_L = jnp.min(state.samples.log_L)
        diff_log_L = jnp.abs(max_log_L - min_log_L)  # NaN = inf - inf
        diff_log_L = jnp.where(jnp.isnan(diff_log_L), jnp.inf, diff_log_L)
        reached_atol = diff_log_L <= term_cond.atol
        done, termination_reason = _set_done_bit(reached_atol, 3,
                                                 done=done, termination_reason=termination_reason)

    if term_cond.min_efficiency is not None:
        # efficiency below threshold
        efficiency = state.samples.log_L.shape[0] / jnp.sum(state.samples.num_likelihood_evaluations)
        reached_min_efficiency = efficiency <= term_cond.min_efficiency
        done, termination_reason = _set_done_bit(reached_min_efficiency, 4,
                                                 done=done, termination_reason=termination_reason)

    # on plateau
    on_plateau = jnp.max(state.samples.log_L) == jnp.min(state.samples.log_L)
    done, termination_reason = _set_done_bit(on_plateau, 5,
                                             done=done, termination_reason=termination_reason)

    return done, termination_reason


def gradient_based_optimisation(model: BaseAbstractModel, init_U_point: UType) -> Tuple[UType, FloatArray, IntArray]:
    def loss(U_unconstrained: UType):
        U = jax.nn.sigmoid(U_unconstrained)
        return -model.log_prob_likelihood(U, allow_nan=False)

    solver = NonlinearCG(
        fun=loss,
        jit=True,
        unroll=False,
        verbose=False
    )

    results = solver.run(init_params=logit(init_U_point))
    return jax.nn.sigmoid(results.params), -results.state.value, results.state.num_fun_eval


def _single_thread_global_optimisation(init_state: GlobalOptimisationState,
                                       termination_cond: GlobalOptimisationTerminationCondition,
                                       sampler: BaseAbstractSampler) -> Tuple[GlobalOptimisationState, IntArray]:
    """
    Runs a single thread of global optimisation. Sequentially samples until termination condition is met,
    replacing the worst sample with a new one in groups of num_samples.

    Args:
        init_state: initial state of the global optimisation
        termination_cond: termination condition
        sampler: sampler

    Returns:
        final state of the global optimisation and termination reason
    """

    class CarryType(NamedTuple):
        state: GlobalOptimisationState

    def cond(carry: CarryType) -> BoolArray:
        done, _ = determine_termination(term_cond=termination_cond, state=carry.state)
        return jnp.bitwise_not(done)

    def body(carry: CarryType) -> CarryType:
        # Devices are independent, i.e. expect no communication between them in sampler.
        key, sample_key = random.split(carry.state.key, 2)

        num_samples = carry.state.samples.log_L.shape[0]

        fake_state = StaticStandardNestedSamplerState(
            key=sample_key,
            next_sample_idx=jnp.asarray(0, int_type),
            sample_collection=StaticStandardSampleCollection(
                sender_node_idx=jnp.zeros((num_samples,), int_type),
                log_L=carry.state.samples.log_L,
                U_samples=carry.state.samples.U_sample,
                num_likelihood_evaluations=carry.state.samples.num_likelihood_evaluations,
                phantom=jnp.zeros((num_samples,), jnp.bool_)
            ),
            front_idx=jnp.arange(num_samples, dtype=int_type)
        )

        k = sampler.num_phantom()
        if k > 0:
            def _repeat(x):
                return jnp.repeat(x, (k + 1), axis=0)

            fake_state = fake_state._replace(
                sample_collection=jax.tree.map(_repeat, fake_state.sample_collection)
            )

        fake_state, fake_termination_register = _inter_sync_shrinkage_process(
            init_state=fake_state,
            sampler=sampler,
            num_samples=num_samples * (1 + k),
            init_termination_register=create_init_termination_register()
        )

        num_likelihood_evaluations = carry.state.num_likelihood_evaluations + jnp.sum(
            fake_state.sample_collection.num_likelihood_evaluations)

        if k > 0:
            # Choose the maximum likelihood sample from the k+1 samples (requires reshape first to unstack)

            choose_idx = jnp.argmax(
                jnp.reshape(fake_state.sample_collection.log_L, ((k + 1), num_samples)),
                axis=0
            )

            def _select(x):
                x = jnp.reshape(x, ((k + 1), num_samples) + x.shape[1:])  # [k+1, N, ...]
                return x[choose_idx, jnp.arange(num_samples)]  # [N, ...]

            fake_state = fake_state._replace(
                sample_collection=jax.tree.map(
                    _select,
                    fake_state.sample_collection
                )
            )

        samples = Sample(
            U_sample=fake_state.sample_collection.U_samples,
            log_L_constraint=fake_state.sample_collection.log_L,
            log_L=fake_state.sample_collection.log_L,
            num_likelihood_evaluations=fake_state.sample_collection.num_likelihood_evaluations
        )

        state = GlobalOptimisationState(
            key=key,
            samples=samples,
            num_likelihood_evaluations=num_likelihood_evaluations,
            num_samples=carry.state.num_samples + jnp.asarray(num_samples, int_type)
        )

        return CarryType(state=state)

    init_carry_state = CarryType(
        state=init_state
    )

    carry_state: CarryType = lax.while_loop(
        cond_fun=cond,
        body_fun=body,
        init_val=init_carry_state
    )

    _, termination_reason = determine_termination(term_cond=termination_cond, state=carry_state.state)

    return carry_state.state, termination_reason


def create_init_state(key: PRNGKey, num_search_chains: int,
                      model: BaseAbstractModel) -> GlobalOptimisationState:
    """
    Creates the initial state of the global optimisation.

    Args:
        key: PRNGKey
        num_search_chains: number of search chains
        model: model

    Returns:
        initial state of the global optimisation
    """
    key, sample_key = random.split(key, 2)
    init_samples = draw_uniform_samples(key=sample_key, num_live_points=num_search_chains, model=model)

    return GlobalOptimisationState(
        key=key,
        samples=init_samples,
        num_likelihood_evaluations=jnp.sum(init_samples.num_likelihood_evaluations),
        num_samples=jnp.asarray(num_search_chains, int_type)
    )


class SimpleGlobalOptimisation:
    """
    Simple global optimisation leveraging building blocks of nested sampling.
    """

    def __init__(self, sampler: BaseAbstractSampler, num_search_chains: int,
                 model: BaseAbstractModel, num_parallel_workers: int = 1):
        self.sampler = sampler
        if num_search_chains < 1:
            raise ValueError("num_search_chains must be >= 1.")
        self.num_search_chains = int(num_search_chains)
        self.num_parallel_workers = int(num_parallel_workers)

        if self.num_parallel_workers > 1:
            logger.info(f"Using {self.num_parallel_workers} parallel workers, each running identical samplers.")
        self.model = model
        self.num_search_chains = num_search_chains

    def _gradient_descent(self, results: GlobalOptimisationResults) -> GlobalOptimisationResults:
        U_solution, log_L_solution, _num_likelihood_evals = gradient_based_optimisation(self.model,
                                                                                        init_U_point=results.U_solution)
        X_solution = self.model.transform(U_solution)
        solution = self.model.prepare_input(U_solution)
        num_likelihood_evals = results.num_likelihood_evaluations + _num_likelihood_evals
        return results._replace(
            U_solution=U_solution,
            log_L_solution=log_L_solution,
            X_solution=X_solution,
            solution=solution,
            num_likelihood_evaluations=num_likelihood_evals
        )

    def _to_results(self, termination_reason: IntArray, state: GlobalOptimisationState) -> GlobalOptimisationResults:
        """
        Converts the final state of the global optimisation to results.

        Args:
            state: final state of the global optimisation

        Returns:
            results of the global optimisation
        """
        best_idx = jnp.argmax(state.samples.log_L)
        U_solution = state.samples.U_sample[best_idx]
        X_solution = self.model.transform(U_solution)
        solution = self.model.prepare_input(U_solution)
        max_log_L = state.samples.log_L[best_idx]
        min_log_L = jnp.min(state.samples.log_L)
        relative_spread = 2. * jnp.abs(max_log_L - min_log_L) / jnp.abs(max_log_L + min_log_L)
        relative_spread = jnp.where(
            jnp.isnan(relative_spread),
            jnp.asarray(jnp.inf, relative_spread.dtype),
            relative_spread
        )
        absolute_spread = jnp.abs(max_log_L - min_log_L)
        absolute_spread = jnp.where(
            jnp.isnan(absolute_spread),
            jnp.asarray(jnp.inf, absolute_spread.dtype),
            absolute_spread
        )
        return GlobalOptimisationResults(
            U_solution=state.samples.U_sample[best_idx],
            X_solution=X_solution,
            solution=solution,
            log_L_solution=state.samples.log_L[best_idx],
            num_likelihood_evaluations=state.num_likelihood_evaluations,
            num_samples=state.num_samples,
            relative_spread=relative_spread,
            absolute_spread=absolute_spread,
            termination_reason=termination_reason
        )

    def _run(self, key: PRNGKey, term_cond: GlobalOptimisationTerminationCondition) -> Tuple[
        IntArray, GlobalOptimisationState]:
        """
        Runs the global optimisation.

        Args:
            key: PRNGKey
            term_cond: termination condition

        Returns:
            the termination reason and final state of the global optimisation
        """

        def replica(key: PRNGKey) -> Tuple[GlobalOptimisationState, IntArray]:
            state = create_init_state(
                key=key,
                num_search_chains=self.num_search_chains,
                model=self.model
            )

            # Continue sampling with provided sampler until user-defined termination condition is met.
            state, termination_reason = _single_thread_global_optimisation(
                init_state=state,
                termination_cond=term_cond,
                sampler=self.sampler
            )
            if self.num_parallel_workers > 1:
                target_log_L_contour = jnp.max(
                    parallel.all_gather(jnp.max(state.samples.log_L), 'i')
                )
                state, termination_reason = _single_thread_global_optimisation(
                    init_state=state,
                    termination_cond=term_cond._replace(log_likelihood_contour=target_log_L_contour),
                    sampler=self.sampler
                )

            return state, termination_reason

        if self.num_parallel_workers > 1:
            parallel_ns = pmap(replica, axis_name='i')
            keys = random.split(key, self.num_parallel_workers)
            batched_state, termination_reason = parallel_ns(keys)
            state = remove_chunk_dim(batched_state)
            state = state._replace(key=state.key[0])
        else:
            state, termination_reason = replica(key)

        return termination_reason, state


def summary(results: GlobalOptimisationResults, f_obj: Optional[Union[str, TextIO]] = None):
    """
    Gives a summary of the results of a global optimisation.

    Args:
        results (GlobalOptimisationResults): Nested sampler result
    """
    main_s = []

    def _print(s):
        print(s)
        main_s.append(s)

    def _round(v, uncert_v):
        v = float(v)
        uncert_v = float(uncert_v)
        try:
            sig_figs = -int("{:e}".format(uncert_v).split('e')[1]) + 1
            return round(float(v), sig_figs)
        except:
            return float(v)

    def _print_termination_condition(_termination_reason: int):
        termination_bit_mask = _bit_mask(int(_termination_reason), width=8)
        # 0-bit -> 1: used maximum allowed number of likelihood evaluations
        #         1-bit -> 2: reached goal log-likelihood contour
        #         2-bit -> 4: relative spread of log-likelihood values below threshold
        #         3-bit -> 8: absolute spread of log-likelihood values below threshold
        #         4-bit -> 16: efficiency below threshold
        #         5-bit -> 32: on a plateau (possibly local minimum, or due to numerical issues)
        for bit, condition in zip(termination_bit_mask, [
            'Reached max num likelihood evaluations',
            'Reached goal log-likelihood contour',
            'Small relative spread of log-likelihood values',
            'Small absolute spread of log-likelihood values',
            'Sampler efficiency too low',
            'On plateau (possibly local minimum, or due to numerical issues)'
        ]):
            if bit == 1:
                _print(condition)

    _print("--------")
    _print("Termination Conditions:")
    if np.size(results.termination_reason) > 1:
        for replica_idx in range(np.size(results.termination_reason)):
            _print(f"Replica {replica_idx}:")
            _print_termination_condition(int(results.termination_reason[replica_idx]))
    else:
        _print_termination_condition(int(results.termination_reason))
    _print("--------")
    _print(f"likelihood evals: {int(results.num_likelihood_evaluations):d}")
    _print(f"samples: {int(results.num_samples):d}")
    _print(
        f"likelihood evals / sample: {float(results.num_likelihood_evaluations / results.num_samples):.1f}"
    )
    _print("--------")
    _print(
        f"max(log_L)={_round(results.log_L_solution, results.log_L_solution)}"
    )
    _print(
        f"relative spread: {_round(results.relative_spread, results.relative_spread)}"
    )
    _print(
        f"absolute spread: {_round(results.absolute_spread, results.absolute_spread)}"
    )

    X_solution = results.X_solution
    for name in X_solution.keys():
        _sample = X_solution[name].reshape((-1,))
        ndims = _sample.size
        _print("--------")
        var_name = name if ndims == 1 else "{}[#]".format(name)
        _print(
            f"{var_name}: max(L) est."
        )
        for dim in range(ndims):
            _max_like_point = _sample[dim]
            # two sig-figs based on uncert
            sig_figs = -int("{:e}".format(_max_like_point * 0.1).split('e')[1]) + 1

            def _round(ar):
                return round(float(ar), sig_figs)

            _print("{}: {}".format(
                name if ndims == 1 else "{}[{}]".format(name, dim),
                _round(_max_like_point)
            ))
    _print("--------")
    if f_obj is not None:
        out = "\n".join(main_s)
        if isinstance(f_obj, str):
            with open(f_obj, 'w') as f:
                f.write(out)
        elif isinstance(f_obj, io.TextIOBase):
            f_obj.write(out)
        else:
            raise TypeError(f"Invalid f_obj: {type(f_obj)}")



import inspect
import warnings
from typing import Tuple, Callable

import jax
from jax import numpy as jnp, lax

from jaxns.framework.bases import PriorModelType, BaseAbstractPrior, PriorModelGen
from jaxns.framework.prior import InvalidPriorName, SingularPrior, Prior
from jaxns.internals.types import UType, XType, float_type, LikelihoodInputType, FloatArray, LikelihoodType, PRNGKey, \
    isinstance_namedtuple

__all__ = [
    'simulate_prior_model'
]


def _get_prior_model_gen(prior_model: PriorModelType) -> PriorModelGen:
    gen = prior_model()
    # Check if gen is a generator
    if not inspect.isgenerator(gen):
        warnings.warn("The provided prior_model is not a generator, this may mean you forget `yield` statements. "
                      "This means there are no Bayesian variables.")

        def dummy_prior_model(output):
            _ = yield Prior(0.)
            return output

        # Make an empty generator that returns the output.
        gen = dummy_prior_model(gen)
    return gen


def compute_U_ndims(prior_model: PriorModelType) -> int:
    """
    Computes placeholders of model.

    Args:
        prior_model: a callable that produces a prior model generator

    Returns:
        number of U dims
    """
    U_ndims = 0
    gen = _get_prior_model_gen(prior_model=prior_model)

    prior_response = None
    names = set()
    while True:
        try:
            prior: BaseAbstractPrior = gen.send(prior_response)
            d = prior.base_ndims
            U_ndims += d
            u = jnp.zeros(prior.base_shape, float_type)
            prior_response = prior.forward(u)
            if prior.name is not None:
                if prior.name in names:
                    raise InvalidPriorName(name=prior.name)
                names.add(prior.name)
        except StopIteration:
            break
    return U_ndims


def simulate_prior_model(key: PRNGKey, prior_model: PriorModelType) -> Tuple[LikelihoodInputType, XType]:
    """
    Simulate a prior model.

    Args:
        key: PRNGKey
        prior_model: A prior model

    Returns:
        a tuple of the likelihood input variables, and dict of non-hidden (named) prior variables.
    """
    U_ndims = compute_U_ndims(prior_model=prior_model)
    U = jax.random.uniform(key, shape=(U_ndims,), dtype=float_type)
    return prepare_input(U=U, prior_model=prior_model), transform(U=U, prior_model=prior_model)


def parse_prior(prior_model: PriorModelType) -> Tuple[UType, XType]:
    """
    Computes placeholders of model.

    Args:
        prior_model: a callable that produces a prior model generator

    Returns:
        U placeholder, X placeholder
    """
    U_ndims = 0
    gen = _get_prior_model_gen(prior_model=prior_model)

    prior_response = None
    names = set()
    X_placeholder: XType = dict()
    while True:
        try:
            prior: BaseAbstractPrior = gen.send(prior_response)
            d = prior.base_ndims
            U_ndims += d
            u = jnp.zeros(prior.base_shape, float_type)
            prior_response = prior.forward(u)
            if prior.name is not None:
                if prior.name in names:
                    raise InvalidPriorName(name=prior.name)
                names.add(prior.name)
                if not isinstance(prior, SingularPrior):
                    X_placeholder[prior.name] = prior_response
        except StopIteration:
            break
    U_placeholder = jnp.zeros((U_ndims,), float_type)
    return U_placeholder, X_placeholder


def parse_joint(prior_model: PriorModelType, log_likelihood: LikelihoodType) -> Tuple[
    UType, XType, LikelihoodInputType, FloatArray]:
    """
    Computes placeholders of model.

    Args:
        prior_model: a callable that produces a prior model generator

    Returns:
        U placeholder, X placeholder
    """
    U_ndims = 0
    gen = _get_prior_model_gen(prior_model=prior_model)

    prior_response = None
    names = set()
    X_placeholder: XType = dict()
    while True:
        try:
            prior: BaseAbstractPrior = gen.send(prior_response)
            d = prior.base_ndims
            U_ndims += d
            u = jnp.zeros(prior.base_shape, float_type)
            prior_response = prior.forward(u)
            if prior.name is not None:
                if prior.name in names:
                    raise InvalidPriorName(name=prior.name)
                names.add(prior.name)
                if not isinstance(prior, SingularPrior):
                    X_placeholder[prior.name] = prior_response
        except StopIteration as e:
            output = e.value
            if (not isinstance(output, tuple)) or isinstance_namedtuple(output):
                output = (output,)
            break
    likelihood_input_placeholder = output
    log_L_placeholder = jnp.asarray(log_likelihood(*output), float_type)
    U_placeholder = jnp.zeros((U_ndims,), float_type)
    return U_placeholder, X_placeholder, likelihood_input_placeholder, log_L_placeholder


def transform(U: UType, prior_model: PriorModelType) -> XType:
    """
    Transforms a flat array of `U_ndims` i.i.d. samples of U[0,1] into the target prior.

    Args:
        U: [U_ndims] a flat array of i.i.d. samples of U[0,1]
        prior_model: a callable that produces a prior model generator

    Returns:
        the prior variables
    """

    gen = _get_prior_model_gen(prior_model=prior_model)

    prior_response = None
    names = set()
    X_collection = dict()
    idx = 0
    while True:
        try:
            prior: BaseAbstractPrior = gen.send(prior_response)
            d = prior.base_ndims
            u = jnp.reshape(U[idx:idx + d], prior.base_shape)
            idx += d
            prior_response = prior.forward(u)
            if prior.name is not None:
                if prior.name in names:
                    raise InvalidPriorName(name=prior.name)
                names.add(prior.name)
                if not isinstance(prior, SingularPrior):
                    X_collection[prior.name] = prior_response
        except StopIteration:
            break
    return X_collection


def transform_parametrised(U: UType, prior_model: PriorModelType) -> XType:
    """
    Transforms a flat array of `U_ndims` i.i.d. samples of U[0,1] into the the parametrised prior variables.

    Args:
        U: [U_ndims] a flat array of i.i.d. samples of U[0,1]
        prior_model: a callable that produces a prior model generator

    Returns:
        the parametrised prior variables
    """

    gen = _get_prior_model_gen(prior_model=prior_model)

    prior_response = None
    names = set()
    Y_collection = dict()
    idx = 0
    while True:
        try:
            prior: BaseAbstractPrior = gen.send(prior_response)
            d = prior.base_ndims
            u = jnp.reshape(U[idx:idx + d], prior.base_shape)
            idx += d
            prior_response = prior.forward(u)
            if prior.name is not None:
                if prior.name in names:
                    raise InvalidPriorName(name=prior.name)
                names.add(prior.name)
                if isinstance(prior, SingularPrior):
                    Y_collection[prior.name] = prior_response
        except StopIteration:
            break
    return Y_collection


def prepare_input(U: UType, prior_model: PriorModelType) -> LikelihoodInputType:
    """
    Transforms a flat array of `U_ndims` i.i.d. samples of U[0,1] into the likelihood conditional variables.

    Args:
        U: [U_ndims] a flat array of i.i.d. samples of U[0,1]
        prior_model: a callable that produces a prior model generator

    Returns:
        the conditional variables of likelihood model
    """

    gen = _get_prior_model_gen(prior_model=prior_model)

    prior_response = None
    idx = 0
    while True:
        try:
            prior: BaseAbstractPrior = gen.send(prior_response)
            d = prior.base_ndims
            u = jnp.reshape(U[idx:idx + d], prior.base_shape)
            idx += d
            prior_response = prior.forward(u)
        except StopIteration as e:
            output = e.value
            if (not isinstance(output, tuple)) or isinstance_namedtuple(output):
                output = (output,)
            break
    return output


def compute_log_prob_prior(U: UType, prior_model: PriorModelType) -> FloatArray:
    """
    Computes the prior log-density from a U-space sample.

    Args:
        U: [U_ndims] a flat array of i.i.d. samples of U[0,1]
        prior_model: a callable that produces a prior model generator

    Returns:
        prior log-density
    """

    gen = _get_prior_model_gen(prior_model=prior_model)

    prior_response = None
    log_prob = []
    idx = 0
    while True:
        try:
            prior: BaseAbstractPrior = gen.send(prior_response)
            d = prior.base_ndims
            u = jnp.reshape(U[idx:idx + d], prior.base_shape)
            idx += d
            prior_response = prior.forward(u)
            log_prob.append(prior.log_prob(prior_response))
        except StopIteration:
            break
    return sum(log_prob, jnp.asarray(0., float_type))


def compute_log_likelihood(U: UType, prior_model: PriorModelType, log_likelihood: LikelihoodType,
                           allow_nan: bool = False) -> FloatArray:
    """
    Computes the log likelihood from U-space sample.

    Args:
        U: [U_ndims] a flat array of i.i.d. samples of U[0,1]
        prior_model: a callable that produces a prior model generator
        log_likelihood: callable that takes arrays returned by the prior model and returns a scalar float
        allow_nan: whether to allow nans in likelihood

    Returns:
        log-likelihood
    """

    V = prepare_input(U=U, prior_model=prior_model)
    log_L = jnp.asarray(log_likelihood(*V), float_type)
    if log_L.size != 1:
        raise ValueError(f"Log likelihood should be scalar, but got {log_L.shape}.")
    if log_L.shape != ():
        log_L = lax.reshape(log_L, ())
    if not allow_nan:
        is_nan = lax.ne(log_L, log_L)
        log_L = lax.select(is_nan, jnp.asarray(-jnp.inf, log_L.dtype), log_L)
    return log_L


def memoize_prior_model(prior_model: PriorModelType, *args, **kwargs) -> Callable:
    """
    Memoize the prior model into a pure function. This can be used, e.g. to compute jacobians, or gradients inside a
    prior model.

    Args:
        prior_model: a prior model
        *args: inputs
        **kwargs: inputs

    Returns:
        a pure function that takes the inputs and passes the prior values appropriately at execution time.
    """
    gen = prior_model(*args, **kwargs)
    prior_response = None
    stack = []
    while True:
        try:
            prior: BaseAbstractPrior = gen.send(prior_response)
            prior_response = yield prior
            stack.append(prior_response)
        except StopIteration as e:
            # output = e.value
            break

    def _pure_fn(*args, **kwargs):
        gen = prior_model(*args, **kwargs)
        stack_iter = iter(stack)
        prior_response = None
        while True:
            try:
                _ = gen.send(prior_response)
                prior_response = next(stack_iter)
            except StopIteration as e:
                output = e.value
                break
        return output

    return _pure_fn



import jax
import jax.random
import numpy as np

from jaxns import Prior, Model
from jaxns.framework.jaxify import jaxify_likelihood
from jaxns.framework.tests.test_model import tfpd


def test_jaxify_likelihood():
    def log_likelihood(x, y):
        return np.sum(x, axis=-1) + np.sum(y, axis=-1)

    wrapped_ll = jaxify_likelihood(log_likelihood)
    np.testing.assert_allclose(wrapped_ll(np.array([1, 2]), np.array([3, 4])), 10)

    vmaped_wrapped_ll = jax.vmap(jaxify_likelihood(log_likelihood, vectorised=True))

    np.testing.assert_allclose(vmaped_wrapped_ll(np.array([[1, 2], [2, 2]]), np.array([[3, 4], [4, 4]])),
                               np.array([10, 12]))


def test_jaxify():
    def prior_model():
        x = yield Prior(tfpd.Uniform(), name='x').parametrised()
        return x

    @jaxify_likelihood
    def log_likelihood(x):
        return x

    model = Model(prior_model=prior_model, log_likelihood=log_likelihood)
    model.sanity_check(key=jax.random.PRNGKey(0), S=10)
    assert model.U_ndims == 0
    assert model.num_params == 1



import json
from typing import NamedTuple

import numpy as np

from jaxns.internals.namedtuple_utils import issubclass_namedtuple, serialise_namedtuple, deserialise_namedtuple


# Example NamedTuple
class MockAge(NamedTuple):
    years: int
    months: np.ndarray


class MockPerson(NamedTuple):
    name: str
    age: MockAge


def test_isinstance_namedtuple():
    # Example NamedTuple
    data = MockPerson('Alice', MockAge(25, np.asarray(6)))
    assert isinstance(data, MockPerson)

    data = ()
    assert not isinstance(data, MockPerson)


def test_issubclass_namedtuple():
    class TestSubclass(NamedTuple):
        pass

    assert issubclass_namedtuple(TestSubclass)

    class TestSubclass(tuple):
        pass

    assert not issubclass_namedtuple(TestSubclass)

    class C:
        pass

    assert not issubclass_namedtuple(C)


def test_serialise_namedtuple():
    class MockAge(NamedTuple):
        years: int
        months: np.ndarray

    class MockPerson(NamedTuple):
        name: str
        age: MockAge

    # Example NamedTuple
    data = MockPerson('Alice', MockAge(25, np.array(6)))
    # Serialise
    serialized_data = serialise_namedtuple(data)
    print(serialized_data)

    # Deserialize
    restored_data = deserialise_namedtuple(serialized_data)
    print(restored_data)

    assert data == restored_data


def test_to_json():
    class MockAge(NamedTuple):
        years: int
        months: np.ndarray

    class MockPerson(NamedTuple):
        name: str
        age: MockAge

    data = MockPerson('Alice', MockAge(25, np.array(6)))
    s = json.dumps(serialise_namedtuple(data), indent=2)
    print(s)
    restored_data = deserialise_namedtuple(json.loads(s))
    print(restored_data)
    assert data == restored_data



import time

import jax
import numpy as np
import pytest
import tensorflow_probability.substrates.jax as tfp

from jaxns import Prior, Model
from jaxns.experimental import EvidenceMaximisation

tfpd = tfp.distributions


@pytest.mark.parametrize("solver", ['adam', 'armijo'])
def test_basic(solver):
    def prior_model():
        x = yield Prior(tfpd.Uniform(0., 1.))
        y = yield Prior(tfpd.Normal(x, 1.), name='y').parametrised()
        sigma = yield Prior(tfpd.Exponential(1.))
        return y, sigma

    def log_likelihood(y, sigma):
        return tfpd.Normal(y, sigma).log_prob(0.)

    model = Model(prior_model=prior_model, log_likelihood=log_likelihood)

    em = EvidenceMaximisation(model=model, ns_kwargs=dict(max_samples=1e5), verbose=False, solver=solver)
    t0 = time.time()
    ns_results, params = em.train(num_steps=3)
    print(f"Time taken ({solver}): {time.time() - t0}")

    def change(x, y):
        if np.size(x) > 0:
            return np.any(np.abs(x - y) > 1e-2)
        return True

    assert all(
        change(p, p_) for p, p_ in zip(jax.tree.leaves(model.params), jax.tree.leaves(params)))


def test_basic_zero_size_param():
    def prior_model():
        x = yield Prior(tfpd.Uniform(0., 1.))
        y = yield Prior(tfpd.Normal(x, 1.), name='y').parametrised()
        z = yield Prior(0., name='z').parametrised()  # This is a zero size parameter
        sigma = yield Prior(tfpd.Exponential(1.))
        return y, z, sigma

    def log_likelihood(y, z, sigma):
        return tfpd.Normal(y, sigma).log_prob(0.) + z

    model = Model(prior_model=prior_model, log_likelihood=log_likelihood)

    em = EvidenceMaximisation(model=model, ns_kwargs=dict(max_samples=1e5))
    assert any(np.size(p) == 0 for p in jax.tree.leaves(model.params))

    ns_results, params = em.train(num_steps=1)

    def change(x, y):
        if np.size(x) > 0:
            return np.any(np.abs(x - y) > 1e-2)
        return True

    assert all(
        change(p, p_) for p, p_ in zip(jax.tree.leaves(model.params), jax.tree.leaves(params)))



from functools import partial
from typing import Tuple, Union, Optional, Literal

import jax
import numpy as np
import tensorflow_probability.substrates.jax as tfp
from jax import numpy as jnp, vmap, lax
from jax._src.scipy.special import gammaln

from jaxns.framework.bases import BaseAbstractPrior
from jaxns.framework.prior import SingularPrior, prior_to_parametrised_singular
from jaxns.internals.log_semiring import cumulative_logsumexp
from jaxns.internals.types import FloatArray, IntArray, BoolArray, float_type, int_type, UType, RandomVariableType, \
    MeasureType

tfpd = tfp.distributions

__all__ = [
    "Bernoulli",
    "Beta",
    "Categorical",
    "ForcedIdentifiability",
    "Poisson",
    "UnnormalisedDirichlet",
    "Empirical",
    "TruncationWrapper"
]


class SpecialPrior(BaseAbstractPrior):
    def parametrised(self, random_init: bool = False) -> SingularPrior:
        """
        Convert this prior into a non-Bayesian parameter, that takes a single value in the model, but still has an associated
        log_prob. The parameter is registered as a `hk.Parameter` with added `_param` name suffix.

        Args:
            random_init: whether to initialise the parameter randomly or at the median of the distribution.

        Returns:
            A singular prior.
        """
        return prior_to_parametrised_singular(self, random_init=random_init)


class Bernoulli(SpecialPrior):
    def __init__(self, *, logits=None, probs=None, name: Optional[str] = None):
        super(Bernoulli, self).__init__(name=name)
        self.dist = tfpd.Bernoulli(logits=logits, probs=probs)

    def _dtype(self):
        return self.dist.dtype

    def _base_shape(self) -> Tuple[int, ...]:
        return self._shape()

    def _shape(self) -> Tuple[int, ...]:
        return tuple(self.dist.batch_shape_tensor()) + tuple(self.dist.event_shape_tensor())

    def _forward(self, U) -> Union[FloatArray, IntArray, BoolArray]:
        return self._quantile(U)

    def _inverse(self, X) -> FloatArray:
        return self.dist.cdf(X)

    def _log_prob(self, X) -> FloatArray:
        return self.dist.log_prob(X)

    def _quantile(self, U):
        probs = self.dist._probs_parameter_no_checks()
        sample = jnp.less(U, probs)
        return sample.astype(self.dtype)


class Beta(SpecialPrior):
    def __init__(self, *, concentration0=None, concentration1=None, name: Optional[str] = None):
        super(Beta, self).__init__(name=name)
        # Special cases for Beta that are faster use the Kumaraswamy distribution
        if isinstance(concentration0, (float, int)) and concentration0 == 1:
            self.dist = tfpd.Kumaraswamy(concentration0=concentration0, concentration1=concentration1)
        elif isinstance(concentration1, (float, int)) and concentration1 == 1:
            self.dist = tfpd.Kumaraswamy(concentration0=concentration0, concentration1=concentration1)
        elif isinstance(concentration0, np.ndarray) and np.all(concentration0 == 1):
            self.dist = tfpd.Kumaraswamy(concentration0=concentration0, concentration1=concentration1)
        elif isinstance(concentration1, np.ndarray) and np.all(concentration1 == 1):
            self.dist = tfpd.Kumaraswamy(concentration0=concentration0, concentration1=concentration1)
        else:
            self.dist = tfpd.Beta(concentration0=concentration0, concentration1=concentration1)

    def _dtype(self):
        return self.dist.dtype

    def _base_shape(self) -> Tuple[int, ...]:
        return self._shape()

    def _shape(self) -> Tuple[int, ...]:
        return tuple(self.dist.batch_shape_tensor()) + tuple(self.dist.event_shape_tensor())

    def _forward(self, U) -> Union[FloatArray, IntArray, BoolArray]:
        return self._quantile(U)

    def _inverse(self, X) -> FloatArray:
        return self.dist.cdf(X)

    def _log_prob(self, X) -> FloatArray:
        return self.dist.log_prob(X)

    def _quantile(self, U):
        X = self.dist.quantile(U)
        return X.astype(self.dtype)


class Categorical(SpecialPrior):
    def __init__(self, parametrisation: Literal['gumbel_max', 'cdf'], *, logits=None, probs=None,
                 name: Optional[str] = None):
        """
        Initialised Categorical special prior.

        Args:
            parametrisation: 'cdf' is good for discrete params with correlation between neighbouring categories,
                otherwise gumbel is better.
            logits: log-prob of each category
            probs: prob of each category
            name: optional name
        """
        super(Categorical, self).__init__(name=name)
        self.dist = tfpd.Categorical(logits=logits, probs=probs)
        self._parametrisation = parametrisation

    def _dtype(self):
        return self.dist.dtype

    def _base_shape(self) -> Tuple[int, ...]:
        if self._parametrisation == 'gumbel_max':
            return self._shape() + (self.dist._num_categories(),)
        elif self._parametrisation == 'cdf':
            return self._shape()

    def _shape(self) -> Tuple[int, ...]:
        return tuple(self.dist.batch_shape_tensor()) + tuple(self.dist.event_shape_tensor())

    def _forward(self, U) -> Union[FloatArray, IntArray, BoolArray]:
        if self._parametrisation == 'gumbel_max':
            return self._quantile_gumbelmax(U)
        elif self._parametrisation == 'cdf':
            return self._quantile_cdf(U)

    def _inverse(self, X) -> FloatArray:
        if self._parametrisation == 'cdf':
            return self.dist.cdf(X)
        raise NotImplementedError()

    def _log_prob(self, X) -> FloatArray:
        return self.dist.log_prob(X)

    def _quantile_gumbelmax(self, U):
        logits = self.dist._logits_parameter_no_checks()
        sample_dtype = self.dtype
        z = -jnp.log(-jnp.log(U))  # gumbel
        draws = jnp.argmax(logits + z, axis=-1).astype(sample_dtype)
        return draws

    def _cdf_gumbelmax(self, X):
        logits = self.dist._logits_parameter_no_checks()
        z = jnp.max(logits, axis=-1, keepdims=True)
        logits -= z
        return jnp.exp(logits) / jnp.sum(jnp.exp(logits), axis=-1, keepdims=True)

    def _quantile_cdf(self, U):
        """
        The quantile for CDF parametrisation.

        Args:
            U: [...]

        Returns:
            [...]
        """
        logits = self.dist._logits_parameter_no_checks()  # [..., N]
        cum_logits = cumulative_logsumexp(logits, axis=-1)  # [..., N]
        cum_logits -= cum_logits[..., -1:]
        N = cum_logits.shape[-1]
        cum_logits_flat = jnp.reshape(cum_logits, (-1, N))
        log_U_flat = jnp.reshape(jnp.log(U), (-1,))

        category_flat = vmap(lambda a, v: jnp.searchsorted(a, v, side='left'))(cum_logits_flat, log_U_flat)
        category = jnp.reshape(category_flat, U.shape)
        return category


class ForcedIdentifiability(SpecialPrior):
    """
    Prior for a sequence of `n` random variables uniformly distributed on U[low, high] such that U[i,...] <= U[i+1,...].
    For broadcasting the resulting random variable is sorted on the first dimension elementwise.

    Args:
        n: number of samples within [low,high]
        low: minimum of distribution
        high: maximum of distribution
        fix_left: if True, the leftmost value is fixed to `low`
        fix_right: if True, the rightmost value is fixed to `high`
    """

    def __init__(self, *, n: int, low=None, high=None, fix_left: bool = False, fix_right: bool = False,
                 name: Optional[str] = None):
        super(ForcedIdentifiability, self).__init__(name=name)
        n_min = (1 if fix_left else 0) + (1 if fix_right else 0)
        if n < n_min:
            raise ValueError(f'`n` too small for fix_left={fix_left} and fix_right={fix_right}')
        self.n = n
        low, high = jnp.broadcast_arrays(low, high)
        self.low = low
        self.high = high
        self.fix_left = fix_left
        self.fix_right = fix_right

    def _dtype(self):
        return float_type

    def _base_shape(self) -> Tuple[int, ...]:
        num_base = self.n
        if self.fix_left:
            num_base -= 1
        if self.fix_right:
            num_base -= 1
        return (num_base,) + np.shape(self.low)

    def _shape(self) -> Tuple[int, ...]:
        return (self.n,) + np.shape(self.low)

    def _forward(self, U) -> Union[FloatArray, IntArray, BoolArray]:
        return self._quantile(U)

    def _inverse(self, X) -> FloatArray:
        return self._cdf(X)

    def _log_prob(self, X) -> FloatArray:
        n = self.n
        if self.fix_left:
            n -= 1
        if self.fix_right:
            n -= 1
        log_n_fac = gammaln(n + 1)
        diff = self.high - self.low
        log_prob = - log_n_fac - n * jnp.log(diff)
        # no check that X is inside high and low
        return log_prob

    def _cdf(self, X):
        if self.fix_left:
            X = X[1:]
        if self.fix_right:
            X = X[:-1]

        n = self.n
        if self.fix_left:
            n -= 1
        if self.fix_right:
            n -= 1

        log_output = jnp.log((X - self.low) / (self.high - self.low))

        # Step 2: Undo cumulative sum operation
        inner = jnp.diff(log_output[::-1], prepend=0, axis=0)[::-1]

        # Step 3: Undo reshaping and division
        k = jnp.arange(n) + 1
        log_x = inner * jnp.reshape(k, (n,) + (1,) * (len(self.shape) - 1))

        # Step 4: Find U by exponentiating log_x
        U = jnp.exp(log_x)

        return U.astype(self.dtype)

    def _quantile(self, U):
        n = self.n
        if self.fix_left:
            n -= 1
        if self.fix_right:
            n -= 1
        log_x = jnp.log(U)  # [n, ...]
        k = jnp.arange(n) + 1
        inner = log_x / lax.reshape(k, (n,) + (1,) * (len(self.shape) - 1))
        log_output = jnp.cumsum(inner[::-1], axis=0)[::-1]
        output = self.low + (self.high - self.low) * jnp.exp(log_output)
        if self.fix_left:
            output = jnp.concatenate([self.low[None], output], axis=0)
        if self.fix_right:
            output = jnp.concatenate([output, self.high[None]], axis=0)
        return output.astype(self.dtype)


def _poisson_quantile_bisection(U, rate, max_iter=15, unroll: bool = True):
    # max_iter is set so that error < 1 up to rate of 1e4
    rate = jnp.maximum(jnp.asarray(rate), 1e-5)
    if np.size(rate) > 1:
        raise ValueError("Rate must be a scalar")

    if np.size(U) > 1:
        U_flat = U.ravel()
        x_final, x_results = vmap(lambda u: _poisson_quantile_bisection(u, rate, max_iter, unroll))(U_flat)
        return x_final.reshape(U.shape), x_results.reshape(U.shape + (max_iter,))

    def smooth_cdf(x, rate):
        return lax.igammac(x + 1., rate)

    def fixed_point_update(x, args):
        (a, b, f_a, f_b) = x

        c = 0.5 * (a + b)
        f_c = smooth_cdf(c, rate)

        left = f_c > U
        a1 = jnp.where(left, a, c)
        f_a1 = jnp.where(left, f_a, f_c)
        b1 = jnp.where(left, c, b)
        f_b1 = jnp.where(left, f_c, f_b)

        a2 = a
        f_a2 = f_a
        b2 = b * 2.
        f_b2 = smooth_cdf(b2, rate)

        bounded = f_b >= U  # a already bounds.

        a = jnp.where(bounded, a1, a2)
        b = jnp.where(bounded, b1, b2)
        f_a = jnp.where(bounded, f_a1, f_a2)
        f_b = jnp.where(bounded, f_b1, f_b2)

        new_x = (a, b, f_a, f_b)

        return new_x, 0.5 * (a + b)

    a = jnp.asarray(0.)
    b = jnp.asarray(rate)
    f_a = jnp.asarray(0.)
    f_b = smooth_cdf(b, rate)
    init = (a, b, f_a, f_b)

    # Dummy array to facilitate using scan for a fixed number of iterations
    (a, b, f_a, f_b), x_results = lax.scan(
        fixed_point_update,
        init,
        jnp.arange(max_iter),
        unroll=max_iter if unroll else 1
    )

    c = 0.5 * (a + b)

    return c, x_results


@partial(jax.jit, static_argnames=("unroll",))
def _poisson_quantile(U, rate, unroll: bool = False):
    x, _ = _poisson_quantile_bisection(U, rate, unroll=unroll)
    return x.astype(int_type)


class Poisson(SpecialPrior):
    def __init__(self, *, rate=None, log_rate=None, name: Optional[str] = None):
        super(Poisson, self).__init__(name=name)
        self.dist = tfpd.Poisson(rate=rate, log_rate=log_rate)

    def _dtype(self):
        return int_type

    def _base_shape(self) -> Tuple[int, ...]:
        return self._shape()

    def _shape(self) -> Tuple[int, ...]:
        return tuple(self.dist.batch_shape_tensor()) + tuple(self.dist.event_shape_tensor())

    def _forward(self, U) -> Union[FloatArray, IntArray, BoolArray]:
        return self._quantile(U)

    def _inverse(self, X) -> FloatArray:
        return self.dist.cdf(X)

    def _log_prob(self, X) -> FloatArray:
        return self.dist.log_prob(X)

    def _quantile(self, U):
        """
        Algorithmic Poisson generator based upon the inversion by sequential search

        .. code-block::

                init:
                    Let log_x ← -inf, log_p ← −λ, log_s ← log_p.
                    Generate uniform random number u in [0,1].
                while log_u > log_s do:
                    log_x ← logaddexp(log_x, 0).
                    log_p ← log_p + log_λ - log_x.
                    log_s ← logaddexp(log_s, log_p).
                return exp(log_x).
        """
        rate = self.dist.rate_parameter()
        if np.size(rate) > 1:
            return jax.vmap(lambda u, r: _poisson_quantile(u, r))(U.ravel(), rate.ravel()).reshape(self.shape)
        return _poisson_quantile(U, rate)


class UnnormalisedDirichlet(SpecialPrior):
    """
    Represents an unnormalised dirichlet distribution of K classes.
    That is, the output is related to the K-simplex via normalisation.

    X ~ UnnormalisedDirichlet(alpha)
    Y = X / sum(X) ==> Y ~ Dirichlet(alpha)
    """

    def __init__(self, *, concentration, name: Optional[str] = None):
        super(UnnormalisedDirichlet, self).__init__(name=name)
        self._dirichlet_dist = tfpd.Dirichlet(concentration=concentration)
        self._gamma_dist = tfpd.Gamma(concentration=concentration, log_rate=0.)

    def _dtype(self):
        return self._dirichlet_dist.dtype

    def _base_shape(self) -> Tuple[int, ...]:
        return self._shape()

    def _shape(self) -> Tuple[int, ...]:
        return tuple(self._dirichlet_dist.batch_shape_tensor()) + tuple(self._dirichlet_dist.event_shape_tensor())

    def _forward(self, U) -> Union[FloatArray, IntArray, BoolArray]:
        return self._quantile(U)

    def _inverse(self, X) -> FloatArray:
        return self._gamma_dist.cdf(X)

    def _log_prob(self, X) -> FloatArray:
        return jnp.sum(self._gamma_dist.log_prob(X), axis=-1)

    def _quantile(self, U):
        gamma = self._gamma_dist.quantile(U).astype(self.dtype)
        return gamma


class Empirical(SpecialPrior):
    """
    Represents the empirical distribution of a set of 1D samples, with arbitrary batch dimension.
    """

    def __init__(self, *, samples: jax.Array, resolution: int = 100, name: Optional[str] = None):
        super(Empirical, self).__init__(name=name)
        if len(np.shape(samples)) < 1:
            raise ValueError("Samples must have at least one dimension")
        if np.size(samples) == 0:
            raise ValueError("Samples must have at least one element")
        if resolution < 1:
            raise ValueError("Resolution must be at least 1")
        samples = jnp.asarray(samples)
        self._q = jnp.linspace(0., 100., resolution + 1)
        self._percentiles = jnp.reshape(jnp.percentile(samples, self._q, axis=-1), (resolution + 1, -1))

        self._batch_shape = np.shape(samples)[:-1]
        self._samples_dtype = samples.dtype

    def _dtype(self):
        return self._samples_dtype

    def _base_shape(self) -> Tuple[int, ...]:
        return self._batch_shape

    def _shape(self) -> Tuple[int, ...]:
        return self._batch_shape

    def _forward(self, U) -> Union[FloatArray, IntArray, BoolArray]:
        return self._quantile(U)

    def _inverse(self, X) -> FloatArray:
        return self._cdf(X)

    def _cdf(self, X) -> FloatArray:
        X_flat = jnp.ravel(X)
        u = jax.vmap(lambda x, per: jnp.interp(x, per, self._q * 1e-2), in_axes=(0, 1))(X_flat, self._percentiles)
        u = lax.reshape(u, X.shape)
        return u

    def _log_prob(self, X) -> FloatArray:
        X_flat = jnp.ravel(X)

        def _cdf(x):
            return self._cdf(x)

        def log_pdf(x):
            return jnp.log(jax.grad(_cdf)(x))

        log_prob = lax.reshape(jax.vmap(log_pdf)(X_flat), X.shape)
        return log_prob

    def _quantile(self, U):
        U_flat = jnp.ravel(U)
        x = jax.vmap(lambda u, per: jnp.interp(u * 100., self._q, per), in_axes=(0, 1))(U_flat, self._percentiles)
        x = lax.reshape(x, U.shape)
        return x


class TruncationWrapper(SpecialPrior):
    """
    Wraps another prior to make it truncated.

    For truncated distribution the quantile transforms to:

        Q_truncated(p) = Q_untruncated( p * (F_truncated(high) - F_truncated(low)) + F_truncated(low))

    And the CDF transforms to:

        F_truncated(x) = (F_untruncated(x) - F_untruncated(low)) / (F_untruncated(high) - F_untruncated(low))
    """

    def __init__(self, prior: BaseAbstractPrior, low: Union[jax.Array, float], high: Union[jax.Array, float],
                 name: Optional[str] = None):
        super(TruncationWrapper, self).__init__(name=name)
        self.prior = prior
        self.low = jnp.minimum(low, high)
        self.high = jnp.maximum(low, high)
        self.cdf_low = self.prior._inverse(self.low)
        self.cdf_diff = self.prior._inverse(self.high) - self.prior._inverse(self.low)

    def _inverse(self, X: RandomVariableType) -> UType:
        return jnp.clip((self.prior._inverse(X) - self.cdf_low) / jnp.maximum(self.cdf_diff, 1e-6),
                        0., 1.)

    def _forward(self, U: UType) -> RandomVariableType:
        return jnp.clip(self.prior._forward(jnp.clip(U * self.cdf_diff + self.cdf_low,
                                                     0., 1.)),
                        self.low, self.high)

    def _log_prob(self, X: RandomVariableType) -> MeasureType:
        outside_mask = jnp.bitwise_or(X < self.low, X > self.high)
        return jnp.where(outside_mask, -jnp.inf, self.prior._log_prob(X) - jnp.log(self.cdf_diff))

    def _base_shape(self) -> Tuple[int, ...]:
        return self.prior._base_shape()

    def _shape(self) -> Tuple[int, ...]:
        return self.prior._shape()

    def _dtype(self) -> jnp.dtype:
        return self.prior._dtype()



from abc import abstractmethod
from typing import Tuple, Optional, Generator, Callable

import jax.numpy as jnp
import numpy as np
from jax import lax

from jaxns.framework.abc import AbstractModel, AbstractPrior, AbstractDistribution
from jaxns.internals.shapes import tuple_prod
from jaxns.internals.types import LikelihoodInputType
from jaxns.internals.types import LikelihoodType, UType, XType, RandomVariableType, MeasureType

__all__ = [
    'BaseAbstractModel',
    "BaseAbstractPrior",
    "BaseAbstractDistribution",
    "PriorModelGen",
    "PriorModelType"
]


class BaseAbstractPrior(AbstractPrior):
    """
    The base prior class with public methods.
    """

    def __init__(self, name: Optional[str] = None):
        self.name = name

    def __repr__(self):
        return f"{self.name if self.name is not None else '*'}\t{self.base_shape} -> {self.shape} {self.dtype}"

    @property
    def dtype(self):
        """
        The dtype of the prior random variable in X-space.
        """
        return self._dtype()

    @property
    def base_shape(self) -> Tuple[int, ...]:
        """
        The base shape of the prior random variable in U-space.
        """
        return self._base_shape()

    @property
    def base_ndims(self):
        """
        The number of dimensions of the prior random variable in U-space.
        """
        return tuple_prod(self.base_shape)

    @property
    def shape(self) -> Tuple[int, ...]:
        """
        The shape of the prior random variable in X-space.
        """
        return self._shape()

    def forward(self, U: UType) -> RandomVariableType:
        """
        The forward transformation from U-space to X-space.

        Args:
            U: U-space representation

        Returns:
            X-space representation
        """
        return self._forward(U)

    def inverse(self, X: RandomVariableType) -> UType:
        """
        The inverse transformation from X-space to U-space.

        Args:
            X: X-space representation

        Returns:
            U-space representation
        """
        return self._inverse(X)

    def log_prob(self, X: RandomVariableType) -> MeasureType:
        """
        The log probability of the prior.

        Args:
            X: X-space representation

        Returns:
            log probability of the prior
        """
        log_prob = self._log_prob(X)
        if np.size(log_prob) > 1:
            log_prob = jnp.sum(log_prob)
        if log_prob.shape != ():
            log_prob = lax.reshape(log_prob, ())
        return log_prob


PriorModelGen = Generator[BaseAbstractPrior, RandomVariableType, LikelihoodInputType]
PriorModelType = Callable[[...], PriorModelGen]


class BaseAbstractModel(AbstractModel):
    """
    The base model class with public methods.
    """

    def __init__(self, prior_model: PriorModelType, log_likelihood: LikelihoodType):
        self._prior_model = prior_model
        self._log_likelihood = log_likelihood

    @property
    def prior_model(self) -> PriorModelType:
        """
        The prior model.
        """
        return self._prior_model

    @property
    def log_likelihood(self) -> LikelihoodType:
        """
        The log likelihood function.

        Returns:
            log likelihood function
        """
        return self._log_likelihood

    @abstractmethod
    def _U_placeholder(self) -> UType:
        ...

    @abstractmethod
    def _X_placeholder(self) -> XType:
        ...

    @property
    def U_placeholder(self) -> UType:
        """
        A placeholder for U-space sample.
        """
        return self._U_placeholder()

    @property
    def X_placeholder(self) -> XType:
        """
        A placeholder for X-space sample.
        """
        return self._X_placeholder()

    @property
    def U_ndims(self) -> int:
        """
        The prior dimensionality.
        """
        return self.U_placeholder.size


# TODO(Joshuaalbert): distribution is too similar to prior, where we only need to be able to extract the log_prob and
#  potential tranformations from the underlying. Perhaps we should just define as ops to create priors. Try making
#  priors just use distribution functionality. Special priors will need treatment.
class BaseAbstractDistribution(AbstractDistribution):
    """
    The base distribution class with public methods.
    """

    @property
    def dtype(self):
        """
        The dtype of the distribution, in X-space.
        """
        return self._dtype()

    @property
    def base_shape(self) -> Tuple[int, ...]:
        """
        The base shape of the distribution, in U-space.
        """
        return self._base_shape()

    @property
    def shape(self) -> Tuple[int, ...]:
        """
        The shape of the distribution, in X-space.
        """
        return self._shape()

    def forward(self, U: UType) -> RandomVariableType:
        """
        The forward transformation from U-space to X-space.

        Args:
            U: U-space representation

        Returns:
            X-space representation
        """
        return self._forward(U)

    def inverse(self, X: RandomVariableType) -> UType:
        """
        The inverse transformation from X-space to U-space.

        Args:
            X: X-space representation

        Returns:
            U-space representation
        """
        return self._inverse(X)

    def log_prob(self, X: RandomVariableType) -> MeasureType:
        """
        The log probability of the distribution.

        Args:
            X: X-space representation

        Returns:
            log probability of the distribution
        """
        return self._log_prob(X)
